import pathlib

from gpt_image.geometry import Geometry
from gpt_image.partition import Partition, PartitionEntryArray
from gpt_image.table import Header, Table


class TableReadError(Exception):
    pass


class Disk:
    """GPT disk

    A disk objects represents a new or existing GPT disk image.  If the file exists,
    it is assumed to be an existing GPT image. If it does not, a new file is created.

    Attributes:
        image_path: file image path (absolute or relative)
    """

    def __init__(self, image_path: str) -> None:
        """Init Disk with a file path"""
        self.image_path = pathlib.Path(image_path)
        self.name = self.image_path.name
        self.sector_size = 512  # 512 bytes

    def open(self):
        """Read existing GPT disk Table"""
        disk_bytes = self.image_path.read_bytes()
        self.size = self.image_path.stat().st_size
        self.geometry = Geometry(self.size, self.sector_size)
        self.table = Table(self.geometry)
        # read the headers
        primary_header_b = disk_bytes[
            self.geometry.primary_header_byte : self.geometry.primary_header_byte
            + self.geometry.header_length
        ]
        backup_header_b = disk_bytes[
            self.geometry.backup_header_byte : self.geometry.backup_header_byte
            + self.geometry.header_length
        ]
        self.table.primary_header = Header(self.geometry)
        self.table.primary_header.read(primary_header_b)
        self.table.secondary_header = Header(self.geometry, is_backup=True)
        self.table.secondary_header.read(backup_header_b)
        # read the partition tables
        primary_part_table_b = disk_bytes[
            self.geometry.primary_array_byte : self.geometry.primary_array_byte
            + self.geometry.array_length
        ]
        backup_part_table_b = disk_bytes[
            self.geometry.backup_array_byte : self.geometry.backup_array_byte
            + self.geometry.array_length
        ]
        if primary_part_table_b != backup_part_table_b:
            raise TableReadError("primary and backup table do not match")
        # unmarshal the partition bytes to objects
        # loop through the entire array and unmarshall the bytes if partition
        # data is found
        for i in range(PartitionEntryArray.EntryCount):
            offset = i * PartitionEntryArray.EntryLength
            partition_bytes = primary_part_table_b[
                offset : offset + PartitionEntryArray.EntryLength
            ]
            new_part = Partition()
            new_part.read(partition_bytes)
            if new_part.type_guid.data != b"\x00" * 16:
                self.table.partitions.entries.append(new_part)

    def create(self, size: int):
        """Create the disk image on Disk

        Creates the basic image structure at the specified path. This zeros
        the disk and writes the protective MBR.

        """
        self.image_path.touch(exist_ok=False)
        self.size = size
        self.geometry = Geometry(self.size, self.sector_size)
        self.table = Table(self.geometry)
        # @TODO: move this into the write method
        with open(self.image_path, "r+b") as f:
            # zero entire disk
            f.write(b"\x00" * self.size)
            f.seek(0)
            f.write(self.table.protective_mbr.as_bytes())
        self.write()

    def write(self):
        """Write the GPT information to disk

        Writes the GPT header and partition tables to disk. Actions that
        happen before this are not written to disk.

        """
        self.table.update()
        with open(self.image_path, "r+b") as f:
            # write primary header
            f.seek(self.geometry.primary_header_byte)
            f.write(self.table.primary_header.as_bytes())

            # write primary partition table
            f.seek(self.geometry.primary_array_byte)
            f.write(self.table.partitions.as_bytes())

            # move to secondary header location and write
            f.seek(self.geometry.backup_header_byte)
            f.write(self.table.secondary_header.as_bytes())

            # write secondary partition table
            f.seek(self.geometry.backup_array_byte)
            f.write(self.table.partitions.as_bytes())

    def write_data(self, data: bytes, partition: Partition, offset: int = 0) -> None:
        # @NOTE: this isn't a GPT function. Writing data should be outside the
        # scope of this module
        """Write data to disk

        Args:
            data: data to write to partition. only bytes supported
            partition: Partition object to write data to
            offset: byte offset for writing data. The default is 0 but can be set to
                support custom offsets
        """
        if not type(data) is bytes:
            raise ValueError(f"data must be of type bytes. found type: {type(data)}")

        with open(self.image_path, "r+b") as f:
            start_lba = int.from_bytes(partition.first_lba.data_bytes, "little")
            start_byte = int(start_lba * self.sector_size)
            with open(self.image_path, "r+b") as f:
                f.seek(start_byte + offset)
                f.write(data)



import uuid

import pytest
from gpt_image.geometry import Geometry
from gpt_image.table import Header, ProtectiveMBR, Table

DISK_SIZE = 2 * 1024 * 1024  # 2 MB disk


@pytest.fixture
def new_geometry():
    return Geometry(DISK_SIZE)


def test_protective_mbr_init(new_geometry):
    geo = new_geometry
    pmbr = ProtectiveMBR(geo)
    assert type(pmbr.as_bytes()) == bytes
    pmbr_bytes = pmbr.as_bytes()
    assert pmbr_bytes[pmbr.partition_type.offset : pmbr.partition_type.end] == b"\xEE"
    assert pmbr_bytes[
        pmbr.start_sector.offset : pmbr.start_sector.end
    ] == geo.primary_header_lba.to_bytes(4, "little")
    assert pmbr_bytes[pmbr.partition_size.offset : pmbr.partition_size.end] == (
        geo.total_sectors - 1
    ).to_bytes(4, "little")


def test_proctective_mbr_read(new_geometry):
    pmbr = ProtectiveMBR(new_geometry)
    # change partition type to test it's read back properly
    pmbr.partition_type.data = b"\xFF"
    pmbr_b = pmbr.as_bytes()
    new_pmbr = ProtectiveMBR(new_geometry)
    new_pmbr.read(pmbr_b)
    assert new_pmbr.partition_type.data == b"\xFF"


def test_header_init_primary(new_geometry):
    geo = new_geometry
    head = Header(geo, uuid.uuid4())
    assert head.backup is False
    assert head.primary_header_lba.data == geo.primary_header_lba
    assert head.secondary_header_lba.data == geo.backup_header_lba
    assert head.partition_array_start.data == geo.primary_array_lba


def test_header_init_backup(new_geometry):
    geo = new_geometry
    head = Header(geo, uuid.uuid4(), is_backup=True)
    assert head.backup is True
    # ensure the header LBA has been swapped
    assert head.primary_header_lba.data == geo.backup_header_lba
    assert head.secondary_header_lba.data == geo.primary_header_lba
    assert head.partition_entry_start_byte == 0
    assert head.header_start_byte == int(32 * geo.sector_size)


def test_header_read(new_geometry):
    head = Header(new_geometry, uuid.uuid4())
    head_b = head.as_bytes()

    head_ex = Header(new_geometry)
    head_ex.read(head_b)
    assert head_ex.header_sig.data == head.header_sig.data
    assert head_ex.revision.data == head.revision.data
    assert head_ex.revision.data == head.revision.data
    assert head_ex.disk_guid.data == head.disk_guid.data


def test_table_init(new_geometry):
    geo = new_geometry
    table = Table(geo)
    assert table.primary_header.backup is False
    assert table.secondary_header.backup is True


def test_checksum_partitions(new_geometry):
    table = Table(new_geometry)
    # partition checksum will be blank before when initialized
    assert table.primary_header.partition_array_crc.data == 0
    table.checksum_partitions(table.primary_header)
    # test that the checksum is no longer zeroed
    assert (
        table.primary_header.partition_array_crc.data == 2874462854
    )  # predictable binascii.crc32


def test_checksum_header(new_geometry):
    table = Table(new_geometry)
    assert table.primary_header.header_crc.data == 0
    table.checksum_header(table.primary_header)
    # @TODO: (sjw) create header with static UUID
    # the crc is not predictable because the Header UUID is dynamic
    assert table.primary_header.header_crc.data > 0


def test_update(new_geometry):
    table = Table(new_geometry)
    table.update()
    assert table.primary_header.header_crc != b"\x00" * 4
    assert table.secondary_header.header_crc != b"\x00" * 4
    assert table.primary_header.partition_array_crc != b"\x00" * 4
    assert table.secondary_header.partition_array_crc != b"\x00" * 4



import uuid
import os

from gpt_image import disk
from gpt_image.partition import Partition


def test_e2e():
    FILE_NAME = "test-disk.raw"
    # cleanup old test file
    if os.path.isfile(FILE_NAME):
        os.remove(FILE_NAME)
    # this should produce a disk that can be opened and validated with gdisk/sgdisk
    new_disk = disk.Disk(FILE_NAME)
    new_disk.create(2 * 1024 * 1024)
    part1 = Partition("test-part", 2 * 1024, uuid.uuid4())

    new_disk.table.partitions.add(part1)
    new_disk.write()



"""
Header information reference: https://en.wikipedia.org/wiki/GUID_Partition_Table

"""
import binascii
import uuid

from gpt_image.entry import Entry
from gpt_image.geometry import Geometry
from gpt_image.partition import PartitionEntryArray


class ProtectiveMBR:
    """Protective MBR Table Entry

    Provides the bare minimum entries needed to represent a protective MBR.
    https://thestarman.pcministry.com/asm/mbr/PartTables.htm#pte

    """

    def __init__(self, geometry: Geometry):
        self._start_padding = Entry(0, 446, 0)  # padding before boot indicator
        self.boot_indicator = Entry(446, 1, 0)  # not bootable
        self.start_chs = Entry(447, 3, 0)  # ignore the start CHS
        self.partition_type = Entry(450, 1, b"\xEE")  # GPT partition type
        self.end_chs = Entry(451, 3, 0)  # ignore the end CHS
        self.start_sector = Entry(454, 4, geometry.primary_header_lba)
        # size, minus the protective MBR sector. This only works if the
        # disk is under 2 TB
        self.partition_size = Entry(458, 4, geometry.total_sectors - 1)
        self._end_padding = Entry(462, 48, 0)  # padding before signature
        self.signature = Entry(510, 2, b"\x55\xAA")

    def as_bytes(self) -> bytes:
        """Get the Protective MBR as bytes"""
        pmbr_fields = [
            self._start_padding,
            self.boot_indicator,
            self.start_chs,
            self.partition_type,
            self.end_chs,
            self.start_sector,
            self.partition_size,
            self._end_padding,
            self.signature,
        ]
        byte_list = [x.data_bytes for x in pmbr_fields]
        return b"".join(byte_list)

    def read(self, pmbr_data: bytes):
        self.boot_indicator.data = pmbr_data[
            self.boot_indicator.offset : self.boot_indicator.end
        ]
        self.start_chs.data = pmbr_data[self.start_chs.offset : self.start_chs.end]
        self.partition_type.data = pmbr_data[
            self.partition_type.offset : self.partition_type.end
        ]
        self.end_chs.data = pmbr_data[self.end_chs.offset : self.end_chs.end]
        self.start_sector.data = pmbr_data[
            self.start_sector.offset : self.start_sector.end
        ]
        self.partition_size.data = pmbr_data[
            self.partition_size.offset : self.partition_size.end
        ]
        self.signature.data = pmbr_data[self.signature.offset : self.signature.end]


class Header:
    """GPT Partition Table Header Object

    Each table has two GPT headers, primary and secondary (backup). The primary is
    written to LBA 1 and secondary is written to LBA -1. The GPT Headers contain
    various data including the locations of one another. Therefore two headers
    are created for each table.  The only common component is the disk GUID.

    """

    def __init__(
        self,
        geometry: Geometry,
        guid: uuid.UUID = uuid.uuid4(),
        is_backup: bool = False,
    ):
        self.geometry = geometry
        self.guid = guid
        self.backup = is_backup

        # all header values start with just offset and length, data is 0
        self.header_sig = Entry(0, 8, b"EFI PART")

        self.revision = Entry(8, 4, b"\x00\x00\x01\x00")

        self.header_size = Entry(12, 4, 92)
        self.header_crc = Entry(16, 4, 0)
        self.reserved = Entry(20, 4, 0)
        self.primary_header_lba = Entry(24, 8, self.geometry.primary_header_lba)

        self.secondary_header_lba = Entry(32, 8, self.geometry.backup_header_lba)
        self.partition_start_lba = Entry(40, 8, self.geometry.partition_start_lba)

        self.partition_last_lba = Entry(48, 8, self.geometry.partition_last_lba)
        self.disk_guid = Entry(56, 16, self.guid.bytes_le)
        self.partition_array_start = Entry(72, 8, self.geometry.primary_array_lba)
        self.partition_array_length = Entry(80, 4, 128)
        self.partition_entry_size = Entry(84, 4, 128)
        self.partition_array_crc = Entry(88, 4, 0)
        self.reserved_padding = Entry(92, 420, 0)

        if self.backup:
            self.primary_header_lba.data, self.secondary_header_lba.data = (
                self.secondary_header_lba.data,
                self.primary_header_lba.data,
            )
            self.partition_array_start.data = (self.geometry.backup_array_lba).to_bytes(
                8, "little"
            )

        # header start byte relative to the table itself, not the disk
        # primary will be 0 secondary will be LBA 32
        self.header_start_byte = 0
        self.partition_entry_start_byte = int(1 * self.geometry.sector_size)
        if self.backup:
            self.header_start_byte = int(32 * self.geometry.sector_size)
            self.partition_entry_start_byte = 0

    def as_bytes(self) -> bytes:
        """Return the header as bytes"""
        header_fields = [
            self.header_sig,
            self.revision,
            self.header_size,
            self.header_crc,
            self.reserved,
            self.primary_header_lba,
            self.secondary_header_lba,
            self.partition_start_lba,
            self.partition_last_lba,
            self.disk_guid,
            self.partition_array_start,
            self.partition_array_length,
            self.partition_entry_size,
            self.partition_array_crc,
        ]
        byte_list = [x.data_bytes for x in header_fields]
        return b"".join(byte_list)

    def read(self, header_bytes) -> None:
        """Unmarshal bytes to Header object"""

        def _to_int(field):
            return int.from_bytes(header_bytes[field.offset : field.end], "little")

        self.header_sig.data = header_bytes[
            self.header_sig.offset : self.header_sig.offset + self.header_sig.length
        ]
        self.revision.data = header_bytes[
            self.revision.offset : self.revision.offset + self.revision.length
        ]
        self.header_size.data = _to_int(self.header_size)
        self.header_crc.data = _to_int(self.header_crc)
        self.reserved.data = _to_int(self.reserved)
        self.primary_header_lba.data = _to_int(self.primary_header_lba)
        self.secondary_header_lba.data = _to_int(self.secondary_header_lba)
        self.partition_start_lba.data = _to_int(self.partition_start_lba)
        self.partition_last_lba.data = _to_int(self.partition_last_lba)
        # stored as little endian bytes. This will need to be converted to display in
        # human readable form
        # uuid.UUID(bytes=self.disk_guid.data)
        self.disk_guid.data = header_bytes[
            self.disk_guid.offset : self.disk_guid.offset + self.disk_guid.length
        ]
        self.partition_array_start.data = _to_int(self.partition_array_start)
        self.partition_array_length.data = _to_int(self.partition_array_length)
        self.partition_entry_size.data = _to_int(self.partition_entry_size)
        self.partition_array_crc.data = _to_int(self.partition_array_crc)


class Table:
    """GPT Partition Table Object

    The entire GPT table structure including the protective MBR,
    GPT headers and partition tables
    """

    def __init__(self, geometry: Geometry):
        self.geometry = geometry
        self.protective_mbr = ProtectiveMBR(self.geometry)
        self.primary_header = Header(self.geometry)
        self.secondary_header = Header(self.geometry, is_backup=True)
        self.partitions = PartitionEntryArray(self.geometry)

    def update(self) -> None:
        # calculate partition checksum and write to header
        self.checksum_partitions(self.primary_header)
        self.checksum_partitions(self.secondary_header)

        # calculate header checksum and write to header
        self.checksum_header(self.primary_header)
        self.checksum_header(self.secondary_header)

    def checksum_partitions(self, header: Header) -> None:
        """Checksum the partition entries

        Args:
            header: initialized GPT header object
        """
        part_entry_bytes = self.partitions.as_bytes()
        header.partition_array_crc.data = binascii.crc32(part_entry_bytes)

    def checksum_header(self, header: Header) -> None:
        """Checksum the table header

        This CRC includes the partition checksum, and must be calculated
        after that has been written.

        Args:
            header: initialized GPT header object
        """
        # zero the old checksum before calculating
        header.header_crc.data = 0
        header.header_crc.data = binascii.crc32(header.as_bytes())
